/*! 
 *********************************************************************
 *  @file   : LDAP_Class.cpp
 *
 *  Project : LDAP Explorer
 *
 *  Package : LDAP
 *
 *  Company : <<>>
 *
 *  Author  : Laurent Ulrich                              Date: 19/11/2002
 *
 *  Purpose : Implementation of methods for class LDAP_Class
 *
 *********************************************************************
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <wx/wxprec.h>
#ifdef __BORLANDC__
  #pragma hdrstop
#endif

#include <wx/msgdlg.h>
#include <wx/config.h>

#if !defined _MSC_VER && !defined __BORLANDC__
#include <sys/time.h>
#endif 

#pragma warning( disable:4503 )

#include <ldap.h>

#include "../include/LDAPBrowser_Defines.h"
#include "../include/LDAP_Class.h"
#include "../include/Utils.h"

char				*LDAP_DefaultAttributes[] = { "+", "*", NULL };
char				*LDAP_Server_TypeNames[] = { "OpenLDAP", "Novell", "Netscape", NULL };
LDAP_SERVER_TYPE	LDAP_Server_Types[] = { OPENLDAP, NOVELL, NETSCAPE, (LDAP_SERVER_TYPE)NULL };


LDAP_Class::LDAP_Class()
{
	m_LDAP = NULL ;
	m_IsMasterCopy = true ;

	m_Results = NULL ;
}

LDAP_Class::~LDAP_Class()
{
	Close();
}


/*!
 * Opens a new connection to an ldap server
 *
 * @param p_Server : server address , name ( can be in the name:port format )
 * @param p_Port : destination port to connect to.
 * @param p_BaseDN : root dn of the tree that we be managed
 * @param p_User : user in the cn=... form
 * @param p_Password : clear text password ( no support for SSL at this time)
 *
 * @return int  : an LDAP success or error code.
 */
int LDAP_Class::Open( const wxString &p_Server, const wxString &p_Port, const wxString &p_BaseDN, const wxString &p_User, const wxString &p_Password, LDAP_SERVER_TYPE p_Type )
{
	ULONG uRet;

	Close();
	
	wxString	Server ;
	long		Port = 389;
	long		Pos ;

	if( ( Pos = p_Server.find( ':' ) ) > 0 )
	{
		Server = p_Server.substr( 0 , Pos );
		Port = atol( p_Server.substr( Pos + 1 , p_Server.length() - ( Pos + 1 ) ).mb_str() );
	}
	else
	{
		Server = p_Server;
	}

	if( p_Port.length() > 0 )
	{
		Port = atol( p_Port.mb_str() );
	}

	m_IsMasterCopy = true;
	m_LDAP = ldap_init( Server.mb_str() , Port );
	if( m_LDAP == NULL )
	{
		return -1;
	}

	/*
	 * lulrich : 28/10/2003 -- these lines are now placed after the ldap_set_option call
	 * due to problem in openldap
	 * uRet = ldap_simple_bind_s( m_LDAP , p_User.mb_str() , p_Password.mb_str() ) ;
	if( uRet != LDAP_SUCCESS )
	{
		return uRet;
	}*/
	int Version = LDAP_VERSION3;
	uRet = ldap_set_option( m_LDAP , LDAP_OPT_PROTOCOL_VERSION ,&Version );
	if( uRet != LDAP_SUCCESS )
	{
		wxString ErrMsg ( ldap_err2string( uRet ), *wxConvCurrent);
		::wxMessageBox( ErrMsg );
		return 0;
	}

	uRet = ldap_simple_bind_s( m_LDAP , p_User.mb_str() , p_Password.mb_str() ) ;
	if( uRet != LDAP_SUCCESS )
	{
		return uRet;
	}
	m_SrvType = p_Type ;
	/*
	int Deref = LDAP_DEREF_NEVER ;
	uRet = ldap_set_option( m_LDAP , LDAP_OPT_DEREF ,&Deref );
	if( uRet != LDAP_SUCCESS )
	{
		::wxMessageBox( ldap_err2string( uRet ) );
		return 0;
	}
	*/
	//wxConfigBase *Conf = wxConfigBase::Get( );
	wxConfig *Conf = new wxConfig( wxT(PACKAGE) );
	if( Conf != NULL )
	{
		int SizeLimit ;
		if( Conf->Read( wxT("SizeLimit"), &SizeLimit ) )
		{	
			uRet = ldap_set_option( m_LDAP , LDAP_OPT_SIZELIMIT	 , &SizeLimit );
		}
	}
	delete Conf;
	return 0;
}

int LDAP_Class::Close()
{
	Free_Results();
	if( m_LDAP != NULL && m_IsMasterCopy )
	{
		ldap_unbind( m_LDAP );
		m_LDAP = NULL;
		m_IsMasterCopy = true;
	}
	return 0;
}


/*!
 * Searches the connected directory and stores internally the results.
 * These results remain usable until the next search op
 *
 * @param *p_BaseDN : base dn of the search . must be a complete dn
 * @param *p_Filter : search filter : ldap search format
 * @param p_GetAttributes : boolean - should we get attributes values
 * @param p_Type : LDAP_SCOPE_SUBTREE - LDAP_SCOPE_ONELEVEL , LDAP_SCOPE_BASE
 *
 * @return int  : LDAP success or error code
 */
int LDAP_Class::Search( const wxString &p_BaseDN , const wxString &p_Filter , int p_GetAttributes , int p_Type = LDAP_SCOPE_SUBTREE, char **p_Attributes )
{
	#define LDAPMESSAGE		m_Results

	BerElement	*Ber=NULL;
	ULONG		uRet;
	timeval		timeout;
	wxString	UTFDn , UTFFilter ;
	int			SizeLimit ;
	char		**Attributes = p_Attributes ;

	timeout.tv_sec = 180;
	timeout.tv_usec = 0;

	Free_Results();
	m_BaseDN = p_BaseDN ;
	m_Type = p_Type ;
	m_Filter = p_Filter;
	m_GetAttributes = p_GetAttributes ;

	if( Attributes == NULL )
	{
		/*
		* If we want to read the schema from openldap some default
		* attributes have to be specified.
		*/
		switch( m_SrvType )
		{
		case OPENLDAP :
			Attributes = LDAP_DefaultAttributes ;
			break;
		default :
			break;
		}
		//Attributes = LDAP_DefaultAttributes ;
	}
	ldap_get_option( m_LDAP, LDAP_OPT_SIZELIMIT, &SizeLimit );
	uRet = ldap_search_ext_s( m_LDAP ,
		p_BaseDN.mb_str() ,					// Level of the search
		p_Type ,					// LDAP_SCOPE_SUBTREE , LDAP_SCOPE_ONELEVEL , LDAP_SCOPE_BASE
		p_Filter.mb_str() ,					// Filter as (objectClass=Phone)
		Attributes ,				// No particular attributes to return
		(p_GetAttributes=0?1:0) ,	// 0=Return Attibutes Values , 1 = Return only attribute names
		NULL ,						// Server controls
		NULL ,						// Client controls
		&timeout ,					// Timeout ...
		SizeLimit ,					// Result Limit
		&LDAPMESSAGE );				// Results ...
	/*
	uRet = ldap_search_ext_s( m_LDAP ,
		NULL ,					// Level of the search
		p_Type ,					// LDAP_SCOPE_SUBTREE , LDAP_SCOPE_ONELEVEL , LDAP_SCOPE_BASE
		p_Filter ,					// Filter as (objectClass=Phone)
		NULL ,						// No particular attributes to return
		(p_GetAttributes=0?1:0) ,	// 0=Return Attibutes Values , 1 = Return only attribute names
		NULL ,						// Server controls
		NULL ,						// Client controls
		&timeout ,					// Timeout ...
		SizeLimit ,					// Result Limit
		&LDAPMESSAGE );				// Results ...
	*/
	if( uRet != LDAP_SUCCESS && uRet != LDAP_SIZELIMIT_EXCEEDED && uRet != LDAP_TIMELIMIT_EXCEEDED)
	{
		if( Ber != NULL )
		{
			ber_free( Ber , 0 );
			Ber = NULL;
		}
		ldap_msgfree( LDAPMESSAGE );
		LDAPMESSAGE = NULL ;
		return uRet;
	}

	return 0;
}


/*!
 * returns the number of results found ( entries in the directory )
 *
 * @param none
 *
 * @return long  : number of entries
 */
unsigned long LDAP_Class::GetSize()
{
	assert( m_Results != NULL );
	return ldap_count_entries( m_LDAP, m_Results );
}


/*!
 * Returns the absolute dn of the specified entry
 *
 * @param p_Index : should be a valid index of the results
 *
 * @return wxString  : dn of the found index or ""
 */
wxString LDAP_Class::GetDN( unsigned long p_Index )
{
	wxString Res = wxT("");

	if( p_Index >= GetSize() )
	{
		return Res;
	}

	LDAPMessage	*Entry = NULL ;
	char		*DN = NULL ;
	Entry = GotoEntry( p_Index );
	if( Entry != NULL )
	{
		DN = ldap_get_dn( m_LDAP, Entry );
		if( DN != NULL )
		{
			Res = wxString( DN, *wxConvCurrent );
			ldap_memfree( DN );
		}
	}
	return Res;
}


/*!
 * returns the number of attributes in the specified entry
 *
 * @param p_ResNumber : index of the wanted entry
 *
 * @return long  : number of attributes
 */
long LDAP_Class::GetAttributeCount( unsigned long p_ResNumber )
{
	unsigned long Nb = 0 ;

	if( p_ResNumber >= GetSize() )
	{
		return 0;
	}

	LDAPMessage *Entry = NULL ;
	char		*AttributeName = NULL;
	BerElement	*Ber = NULL ;	
	

	Entry = GotoEntry( p_ResNumber ) ;
	if( Entry != NULL )
	{
		for( AttributeName = ldap_first_attribute( m_LDAP, Entry, &Ber );
		AttributeName != NULL; 
		AttributeName = ldap_next_attribute( m_LDAP, Entry, Ber ) ) 
		{
			ldap_memfree( AttributeName );
			Nb ++ ;
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber, 0 );
		}

	}

	return Nb;
}


/*!
 * retreives the specified attributes ,of the specified entry , values ...
 *
 * @param p_ResIndex : wanted search result index
 * @param p_Index : wanted attribute index
 * @param &p_Name : out / name of the attribute
 * @param &p_Values : out / values of the attribute
 *
 * @return int  : 0
 */
int LDAP_Class::GetAttribute( unsigned long p_ResIndex, unsigned long p_Index, wxString &p_Name, StringVector &p_Values )
{
	if( p_ResIndex >= GetSize() )
	{
		return 0;
	}
	LDAPMessage 	*Entry = NULL ;
	char			*AttributeName = NULL;
	BerElement		*Ber = NULL ;	
	char			**Values = NULL ;
	berval			**BValues = NULL ;
	unsigned long	i = 0 ;
	

	Entry = GotoEntry( p_ResIndex ) ;
	if( Entry != NULL )
	{
		for( AttributeName = ldap_first_attribute( m_LDAP, Entry, &Ber );
		AttributeName != NULL && i <= p_Index ; 
		AttributeName = ldap_next_attribute( m_LDAP, Entry, Ber ) ) 
		{
			if( i == p_Index )
			{
				p_Name = wxString( AttributeName, *wxConvCurrent );
				
				Values = ldap_get_values( m_LDAP, Entry, AttributeName );
				if( Values != NULL )
				{
					for( long j = 0 ; Values[ j ] != NULL ; j ++ )
					{
						/*
						* Binary values appear here as non UTF-8 - NULL terminated buffers
						* we first test if value is UTF-8
						*/
						if( IsSafeChar( (const unsigned char *)Values[ j ], 
							strlen( Values[ j ] ), true ) )
						{
							/*
							* Values are UTF8 encoded . We must first decode them
							* to UNICODE and the to the current OS Charset (wxConvCurrent)
							*/
							wxString MBString( wxConvUTF8.cMB2WC( Values[ j ] ), *wxConvCurrent );
							p_Values.insert( p_Values.end(), MBString.c_str() ); 
						}
						else
						{
							/*
							* Value is not UTF-8 : we are going to change the values to 
							* make it printable
							*/

							wxString Printable ;
							BValues = ldap_get_values_len( m_LDAP, Entry, AttributeName );
							if( BValues != NULL && BValues[ j ] != NULL )
							{
								ToBinaryPrint( BValues[ j ]->bv_val, BValues[ j ]->bv_len, 
									Printable );
							}
							else
							{
								ToBinaryPrint( Values[ j ], strlen( Values[ j ] ), Printable );
							}
							if( BValues != NULL )
							{
								ldap_value_free_len( BValues );
								BValues = NULL ;
							}

							p_Values.insert( p_Values.end(), Printable ); 
						}
					}
					ldap_value_free( Values );
				}
				if( BValues != NULL )
				{
					ldap_value_free_len( BValues );
				}
				
			}
			ldap_memfree( AttributeName );
			i ++ ;
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber, 0 );
		}

	}
	return 0;
}


/*!
 * effectively tries to write the entry in the ldap directory
 *
 * @param &p_Object : objet to write
 * @param p_IsNew : ldap_add (true) or ldap_modify (false)
 *
 * @return int  : ldap success or error code
 */

int LDAP_Class::WriteMods( const wxString &p_DN, CSTL_ModsVector & p_Mods, bool p_IsNew )
{
	long						AttributeNumber;
	LDAPMod						**Attributes;
	CSTL_ModsVector::iterator	iter;
	long						i = 0 ;
	int							iRet;

	AttributeNumber = p_Mods.size() + 1;

	Attributes = new LDAPMod *[ AttributeNumber ];
	if( Attributes == NULL )
	{
		return -1;
	}

	for( iter = p_Mods.begin(); iter != p_Mods.end(); iter ++ )
	{
		Attributes[ i ] = p_Mods[ i ] ;
		i++;
	}
	Attributes[ i ] = NULL;

	if( p_IsNew )
	{
		iRet = ldap_add_s( m_LDAP, p_DN.mb_str(), Attributes );
	}
	else
	{
		iRet = ldap_modify_s( m_LDAP, p_DN.mb_str(), Attributes );
		if( iRet != LDAP_SUCCESS )
		{
		}
	}
	if( iRet == LDAP_SUCCESS )
	{
		Search( m_BaseDN.c_str(), m_Filter.c_str(), m_GetAttributes, m_Type );
	}
	delete []Attributes ;
	return iRet;
}


/*!
 * tries to delete the specified dn
 *
 * @param *p_DN : dn of the entry to delete
 *
 * @return int  : ldap result code
 */
int LDAP_Class::Delete( const wxString &p_DN )
{
	int	iRet;
	wxString Value = p_DN ;
//	wxString UTF8Value( wxConvUTF8.cWC2MB( Value.wc_str( *wxConvCurrent ) ) );

	iRet = ldap_delete_s( m_LDAP , Value.mb_str( wxConvUTF8 ) );
	return iRet;
}


/*!
 * tries to change the dn of an entry
 *
 * @param *p_OldDN : the entry to change
 * @param *p_NewDN : the entry's new dn
 *
 * @return int  : ldap result code
 */
int LDAP_Class::Rename( const wxString &p_OldDN, const wxString &p_NewDN, bool p_Move )
{
	int iRet;
	
	char **NewElts = NULL;
//	wxString OldDN = p_OldDN ;
//	wxString NewDN = p_NewDN ;
//	wxString UTF8Value1( wxConvUTF8.cWC2MB( OldDN.wc_str( *wxConvCurrent ) ) );
//	wxString UTF8Value2( wxConvUTF8.cWC2MB( NewDN.wc_str( *wxConvCurrent ) ) );

	NewElts = ldap_explode_dn( p_NewDN.mb_str( wxConvUTF8 ) , 0 );
	if( NewElts == NULL)
		return -1;
	wxString NewParent ;

	long nb = ldap_count_values( NewElts );

	for( long i = 1 ; i <  nb ; i++ )
	{
		NewParent += wxString( NewElts[i], *wxConvCurrent );
		if( i != ( nb - 1 ) )
		{
			NewParent += wxT(",") ;
		}
	}
		//iRet = ldap_rename_s( m_LDAP , UTF8Value1.c_str() ,  NewElts[ 0 ] , NewParent.c_str() , 0 , NULL , NULL );
	if( p_Move )
	{
		iRet = ldap_rename_s( m_LDAP, p_OldDN.mb_str( wxConvUTF8 ), NewElts[ 0 ], NewParent.mb_str( wxConvUTF8 ), 1, NULL, NULL );
		//iRet = ldap_rename_s( m_LDAP, p_OldDN, NewElts[ 0 ], NewParent.c_str(), 0, NULL, NULL );
	}
	else
	{
		iRet = ldap_rename_s( m_LDAP, p_OldDN.mb_str( wxConvUTF8 ), NewElts[ 0 ], NULL, 1, NULL, NULL );
		//iRet = ldap_rename_s( m_LDAP, p_OldDN, NewElts[ 0 ], NULL, 1, NULL, NULL );
	}
	ldap_value_free( NewElts );
	return iRet;
}






LDAP * LDAP_Class::GetLDAPHandle()
{
	return m_LDAP ;
}


/*!
 * return the first value of the attribute p_Name if it exists in first resultset
 *
 * @param &p_Name : name of the attribute to retrieve
 *
 * @return wxString  : empty if not found
 */
wxString LDAP_Class::GetAttributeValue( const wxString &p_Name )
{
	wxString Ret;

	LDAPMessage *Entry = NULL ;
	BerElement	*Ber = NULL ;	
	char		**Values = NULL ;
	berval		**BValues = NULL ;
	char		*AttributeName ;
	bool		Found = false;	

	Entry = GotoEntry( 0 ) ;
	if( Entry != NULL )
	{
		for ( AttributeName = ldap_first_attribute( m_LDAP , Entry , &Ber );
		AttributeName != NULL && !Found ; 
		AttributeName = ldap_next_attribute( m_LDAP , Entry , Ber ) ) 
		{
			if( stricmp( p_Name.mb_str(), AttributeName ) == 0 )
			{
				Found = true;
				Values = ldap_get_values( m_LDAP , Entry , AttributeName );
				if( Values != NULL && Values[ 0 ] != NULL )
				{
						if( IsSafeChar( (const unsigned char *)Values[ 0 ], 
							strlen( Values[ 0 ] ), true ) )
						{
							/*
							* Values are UTF8 encoded . We must first decode them
							* to UNICODE and the to the current OS Charset (wxConvCurrent)
							*/
							wxString MBString( wxConvUTF8.cMB2WC( Values[ 0 ] ), *wxConvCurrent );
							Ret = MBString.c_str(); 
						}
						else
						{
							/*
							* Value is not UTF-8 : we are going to change the values to 
							* make it printable
							*/

							wxString	Printable ;
							BValues = ldap_get_values_len( m_LDAP, Entry, AttributeName );
							if( BValues != NULL && BValues[ 0 ] != NULL )
							{
								ToBinaryPrint( BValues[ 0 ]->bv_val, BValues[ 0 ]->bv_len, 
									Printable );
							}
							else
							{
								ToBinaryPrint( Values[ 0 ], strlen( Values[ 0 ] ), Printable );
							}
							Ret = Printable ;
						}
				}
				if( Values != NULL )
				{
					ldap_value_free( Values );
				}
				if( BValues != NULL )
				{
					ldap_value_free_len( BValues );
				}
			}
			ldap_memfree( AttributeName );
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber , 0 );
		}
	}
	return Ret;
}

berval* LDAP_Class::GetAttributeBinValue( const wxString &p_Name, long p_Index )
{
	LDAPMessage *Entry = NULL ;
	BerElement	*Ber = NULL ;	
	berval		**Values = NULL ;
	char		*AttributeName ;
	bool		Found = false;	
	berval		*Ret = NULL;

	Entry = GotoEntry( 0 ) ;
	if( Entry != NULL )
	{
		for ( AttributeName = ldap_first_attribute( m_LDAP , Entry , &Ber );
		AttributeName != NULL && !Found ; 
		AttributeName = ldap_next_attribute( m_LDAP , Entry , Ber ) ) 
		{
			if( stricmp( p_Name.mb_str(), AttributeName ) == 0 )
			{
				Found = true;
				
				Values = ldap_get_values_len( m_LDAP , Entry , AttributeName );
				
				if( Values != NULL && ldap_count_values_len( Values ) >= p_Index )
				{
					Ret = new berval ;
					if( Ret != NULL )
					{
						Ret->bv_len = Values[ p_Index ]->bv_len ;
						Ret->bv_val = new char[ Values[ p_Index ]->bv_len ];
						if( Ret->bv_val != NULL )
						{
							memcpy( Ret->bv_val, Values[ p_Index ]->bv_val, Values[ 0 ]->bv_len );
							Ret->bv_len = Values[ p_Index ]->bv_len ;
						}
					}
				}
				if( Values != NULL )
				{
					ldap_value_free_len( Values );
				}
			}
			ldap_memfree( AttributeName );
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber , 0 );
		}
	}
	return Ret;
}

berval** LDAP_Class::GetAttributeBinValues( long p_ResIndex, long p_AttIndex, wxString &p_Name )
{
	LDAPMessage *Entry = NULL ;
	BerElement	*Ber = NULL ;	
	berval		**Values = NULL ;
	char		*AttributeName ;
	bool		Found = false;	
	berval		**Ret = NULL;
	long		i = 0;

	Entry = GotoEntry( p_ResIndex ) ;
	if( Entry != NULL )
	{
		for ( AttributeName = ldap_first_attribute( m_LDAP , Entry , &Ber );
		AttributeName != NULL && !Found ; 
		AttributeName = ldap_next_attribute( m_LDAP , Entry , Ber ) ) 
		{
			if( i == p_AttIndex )
			{
				Found = true;
				p_Name = wxString( AttributeName, *wxConvCurrent );
				Values = ldap_get_values_len( m_LDAP , Entry , AttributeName );
				Ret = Values;
			}
			ldap_memfree( AttributeName );
			i++;
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber , 0 );
		}
	}
	return Ret;
}
/*!
 * return all the values of the attribute p_Name if it exists
 *
 * @param &p_Name : name of the attribute
 *
 * @return StringVector  : empty if not found
 */
StringVector LDAP_Class::GetAttributeValues( const wxString &p_Name )
{
	StringVector	Ret;
	LDAPMessage *Entry = NULL ;
	char		*AttributeName = NULL;
	BerElement	*Ber = NULL ;	
	char		**Values = NULL ;
	berval		**BValues = NULL;
	bool		Found = false;
	

	Entry = GotoEntry( 0 ) ;
	if( Entry != NULL )
	{
		for ( AttributeName = ldap_first_attribute( m_LDAP , Entry , &Ber );
			AttributeName != NULL && !Found ; 
			AttributeName = ldap_next_attribute( m_LDAP , Entry , Ber ) ) 
		{
			if( stricmp( p_Name.mb_str(), AttributeName ) == 0 )
			{
				Found = true;
				Values = ldap_get_values( m_LDAP , Entry , AttributeName );

				if( Values != NULL )
				{
					for( long j = 0 ; Values[ j ] != NULL ; j ++ )
					{
						/*
						* Binary values appear here as non UTF-8 - NULL terminated buffers
						* we first test if value is UTF-8
						*/
						if( IsSafeChar( (const unsigned char *)Values[ j ], 
							strlen( Values[ j ] ), true ) )
						{
							/*
							* Values are UTF8 encoded . We must first decode them
							* to UNICODE and the to the current OS Charset (wxConvCurrent)
							*/
							//wxString MBString( wxConvUTF8.cMB2WC( Values[ j ] ), *wxConvCurrent );
							wxString MBString( Values[ j ], *wxConvCurrent );
							Ret.insert( Ret.end(), MBString.c_str() );
						}
						else
						{
							/*
							* Value is not UTF-8 : we are going to change the values to 
							* make it printable
							*/
							wxString Printable ;
							BValues = ldap_get_values_len( m_LDAP, Entry, AttributeName );
							if( BValues != NULL && BValues[ j ] != NULL )
							{
								ToBinaryPrint( BValues[ j ]->bv_val, BValues[ j ]->bv_len, 
									Printable );
							}
							else
							{
								ToBinaryPrint( Values[ j ], strlen( Values[ j ] ), Printable );
							}
							if( BValues != NULL )
							{
								ldap_value_free_len( BValues );
								BValues = NULL ;
							}

							Ret.insert( Ret.end(), Printable ); 
						}
					}
					ldap_value_free( Values );
				}
			}
			ldap_memfree( AttributeName );
		}
		if( Ber != NULL ) 
		{
			ber_free( Ber , 0 );
		}
	}
	return Ret;
}

void LDAP_Class::Free_Results()
{
	if( m_Results != NULL )
	{
		ldap_msgfree( m_Results );
		m_Results = NULL ;
	}
}

LDAPMessage * LDAP_Class::GotoEntry( long p_Index )
{
	LDAPMessage	*Entry = NULL ;
	long		i = 0 ;
	
	for ( Entry = ldap_first_entry( m_LDAP , m_Results ); 
		Entry != NULL && i < p_Index  ; 
		Entry = ldap_next_entry( m_LDAP , Entry ) )
	{
		i ++ ;
	}
	return Entry ;
}

LDAPMessage * LDAP_Class::GetEntry( long p_Index )
{
	return GotoEntry( p_Index );
}

int LDAP_Class::Open( LDAP *p_LDAP, LDAP_SERVER_TYPE p_Type )
{
	Close();

	m_IsMasterCopy = false;
	m_SrvType = p_Type ;
	m_LDAP = p_LDAP ;
	return 0;
}


void LDAP_Class::SetSizeLimit( int &p_SizeLimit )
{
	if( m_LDAP == NULL )
	{
		return;
	}
	ldap_set_option( m_LDAP , LDAP_OPT_SIZELIMIT, &p_SizeLimit );
}

int LDAP_Class::GetSizeLimit()
{
	int SizeLimit;
	if( m_LDAP == NULL )
	{
		return 0;
	}
	ldap_get_option( m_LDAP , LDAP_OPT_SIZELIMIT, &SizeLimit );
	return SizeLimit;

}

LDAP_SERVER_TYPE LDAP_Class::GetSrvType()
{
	return m_SrvType ;
}
