/*! 
 *********************************************************************
 *  @file   : LDIF_File.cpp
 *
 *  Project : LDAP Explorer
 *
 *  Package : LDAP
 *
 *  Company : <<>>
 *
 *  Author  : Laurent Ulrich                          Date: 19/03/2003
 *
 *  Purpose : Implementation of methods for class LDIF_File
 *
 *********************************************************************
 */

#include <wx/wxprec.h>
#ifdef __BORLANDC__
  #pragma hdrstop
#endif

#include <wx/msgdlg.h>
#include <wx/intl.h>

#include <ldap.h>

#include "../include/LDAPBrowser_Defines.h"
#include "../include/LDAP_Class.h"
#include "../include/LDIF_File.h"
#include "../include/Utils.h"

char	Base64Table[]	= "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
char	Pad64			= '=';

LDIF_Record::LDIF_Record()
{
	Init();
}
LDIF_Record::~LDIF_Record()
{
}


LDIF_File::LDIF_File() 
{
	m_File = NULL ;
	m_LDAP = NULL ;
}

LDIF_File::~LDIF_File()
{
	Close();
}

int LDIF_File::OpenFile( const wxString & p_FileName )
{
	m_File = fopen( p_FileName.mb_str(), "r" );
	if( m_File == NULL )
	{
		return LDIF_FILE_ERROR;
	}

	fseek( m_File, 0, SEEK_SET );

	return LDIF_NO_ERROR;
}

int LDIF_File::Close()
{
	if( m_File != NULL )
	{
		fflush( m_File );
		fclose( m_File );
		m_File = NULL;
	}
	return LDIF_NO_ERROR;
}

int LDIF_File::ReadHeader()
{
	assert( m_File != NULL );
	int	Len;

	
	fseek( m_File, 0, SEEK_SET );

	Len = ReadLine();
	if( Len <= 0 )
	{
		return Len;
	}

	if( strncmp( m_Line.c_str(), "version:", 8 ) != 0 ) 
	{
		// The file should begin with a version: tag
		return LDIF_BAD_FORMAT;
	}
	
	// read the version 
	long i = 8 ;
	while( m_Line[ i ] != LDIF_LF && m_Line[ i ] == LDIF_SPACE )
	{
		i++;
	}

	// the version MUST be 1
	if( m_Line[i] != '1' )
	{
		return LDIF_BAD_VERSION;
	}
	return LDIF_NO_ERROR;
}

int LDIF_File::ReadLine()
{
	assert( m_File != NULL );
	char	c = 0;
	bool	Continue = true;
	bool	CheckSpace = false;

	m_Line.Reset();

	while( Continue )
	{
		if( feof( m_File ) != 0 )
		{
			Continue = false;
		}
		else if( CheckSpace )
		{
			c = getc( m_File );
			if( c != ' ' )
			{
				Continue = false ;
				ungetc( c, m_File );
				
			}
			CheckSpace = false;
		}
		else
		{
			c = getc( m_File );
			if( c == LDIF_CR )
			{
				// ignore
			}
			else if( c == LDIF_LF && m_Line.length() > 0 )
			{
				CheckSpace = true;
			}
			else if( c == LDIF_LF && m_Line.length() == 0 )
			{
				/*
				* This is a simple LDIF_LF line just ignore it
				*/
			}
			else
			{
				m_Line.append( c );
			}
		}
	}
	if( feof( m_File ) && m_Line.length() == 0 )
	{
		return LDIF_EOF;
	}

	return m_Line.length();
}

bool LDIF_File::Base64Decode( const DynamicBuffer &p_Input, bool & p_IsBinary, 
							 DynamicBuffer &p_Ret )
{
	char			c ;
	short			OctetNumber = 0 ;
	long			Index = 0 ;
	char			*PosChar = NULL ;
	char			CurrentChar = '\0';

	//p_Size = 0 ;
	p_IsBinary = false;


	c = p_Input[ Index ] ;
	while( c != '\0' ) 
	{
		if( c == Pad64 )
		{
			// End of the Base64 string
			break;
		}

		PosChar = strchr( Base64Table, c );
		if( PosChar == 0 )
		{
			// the current char is not a base64 char
			// don't take this as an error, just pass 
			continue;
		}

		// Get the Number associated whith the character
		c = ( PosChar - Base64Table ) ;
		
		if( OctetNumber == 0 )
		{
			
			OctetNumber ++ ;
			// Begin of the first octet
			CurrentChar = ( c << 2 ) ;
		}
		else if( OctetNumber == 1 )
		{
			OctetNumber ++ ;

			CurrentChar = CurrentChar | ( c >> 4 ) ;
			// First Octet decoded
			p_Ret.append( CurrentChar );
			//p_Size ++ ;

			// Begin of the second octet
			CurrentChar = ( c & 0x0F ) << 4 ;

		}
		else if( OctetNumber == 2 )
		{
			OctetNumber ++ ;
			CurrentChar = CurrentChar | ( c >> 2 ) ;
			// Second Octet decoded 
			p_Ret.append( CurrentChar );
			//p_Size ++ ;
			// Begin of the third octet
			CurrentChar = ( c & 0x03 ) << 6 ;
		}
		else if( OctetNumber == 3 )
		{
			OctetNumber = 0 ;
			
			CurrentChar = CurrentChar | c ;
			// Third octet decoded
			p_Ret.append( CurrentChar );
			//p_Size ++ ;
		}
		Index ++ ;
		c = p_Input[ Index ] ;

	}

	if( p_Ret.find( '\0' ) < p_Ret.length() )
	{
		p_IsBinary = true;
	}
	return true;
}

bool LDIF_File::Base64Encode( DynamicBuffer &p_Ret , const unsigned char *p_Buffer
							 , unsigned long p_Size )
{
	long			TripletNumber;
	long			Reliqua;
	unsigned char	Out[4] ;
	unsigned char	In[3] ;

	p_Ret.Reset();
	
	TripletNumber = p_Size / 3 ;
	Reliqua = p_Size % 3 ;

	for( long i = 0 ; i < TripletNumber ; i ++ )
	{

		long Pos = i * 3 ;
		
		In[ 0 ] = p_Buffer[ Pos ];
		In[ 1 ] = p_Buffer[ Pos + 1 ];
		In[ 2 ] = p_Buffer[ Pos + 2 ];

		Out[ 0 ] = In[ 0 ] >> 2;
		Out[ 1 ] = ( ( In[ 0 ] & 0x03 ) << 4) + ( In[ 1 ] >> 4 );
		Out[ 2 ] = ( ( In[ 1 ] & 0x0F ) << 2 ) + ( In[ 2 ] >> 6 );
		Out[ 3 ] = In[ 2 ] & 0x3F;
		

		p_Ret.append( Base64Table[ Out[ 0 ] ] );
		p_Ret.append( Base64Table[ Out[ 1 ] ] );
		p_Ret.append( Base64Table[ Out[ 2 ] ] );
		p_Ret.append( Base64Table[ Out[ 3 ] ] );
	}
	In[ 0 ] = In[ 1 ] = In[ 2 ] = 0 ;
	
	if( Reliqua > 0 )
	{
		In[ 0 ] = p_Buffer[ TripletNumber * 3 + 1 ]  ;
		if( Reliqua == 2 )
		{
			In[ 1 ] = p_Buffer[ TripletNumber * 3 + 2 ] ;
		}
		Out[ 0 ] = In[ 0 ] >> 2;
		Out[ 1 ] = ( ( In[ 0 ] & 0x03 ) << 4) + ( In[ 1 ] >> 4 );
		Out[ 2 ] = ( ( In[ 1 ] & 0x0F ) << 2 ) ;
		
		p_Ret.append( Base64Table[ Out[ 0 ] ] );
		p_Ret.append( Base64Table[ Out[ 1 ] ] );

		if( Reliqua == 2 )
		{
			p_Ret.append( Base64Table[ Out[ 2 ] ] );
		}
		else
		{
			p_Ret.append( Pad64 );
		}
		p_Ret.append( Pad64 );
	}

	return true;
}


//DEL int LDIF_File::AnalyseLine( const ByteArray &p_Line, LDIF_Record & p_Record )
//DEL {
//DEL 	long 			Pos = p_Line.find( ':' );
//DEL 	ByteArray		Left, Right;
//DEL 	unsigned long	Size = 0;
//DEL 	bool			IsBinary = false ;
//DEL 
//DEL 	if( Pos == -1 )
//DEL 	{
//DEL 		return LDIF_BAD_LINE ;
//DEL 	}
//DEL 
//DEL 	Left = p_Line.substr( 0, Pos );
//DEL 	Right = p_Line.substr( Pos + 1 );
//DEL 
//DEL 	if( Right[0] == ':' )
//DEL 	{
//DEL 		// Value is Base64 encoded
//DEL 		Right.erase( 0, 1 );
//DEL 		while( Right[ 0 ] == ' ' && Right.length() > 0 )
//DEL 		{
//DEL 			Right.erase( 0, 1 );
//DEL 		}
//DEL 		Right = Right.substr( 1 );
//DEL 		Right = Base64Decode( Right, Size, IsBinary );
//DEL 		if( IsBinary )
//DEL 		{
//DEL 			berval *Ber = NULL ;
//DEL 			Ber = new berval ;
//DEL 			if( Ber == NULL )
//DEL 			{
//DEL 				return LDIF_MEMORY_ERROR ;
//DEL 			}
//DEL 			Ber->bv_val = new char[ Size ];
//DEL 			memcpy( Ber->bv_val, Right.c_str(), Size );
//DEL 			Ber->bv_len = Size ;
//DEL 			p_Record.m_BerAttributes[ Left ].push_back( Ber );
//DEL 		}
//DEL 		else
//DEL 		{
//DEL 			p_Record.m_Attributes[ Left ].push_back( Right );
//DEL 		}
//DEL 	}
//DEL 	else
//DEL 	{
//DEL 		while( Right[ 0 ] == ' ' && Right.length() > 0 )
//DEL 		{
//DEL 			Right.erase( 0, 1 );
//DEL 		}
//DEL 		p_Record.m_Attributes[ Left ].push_back( Right );
//DEL 	}
//DEL 	
//DEL 	return LDIF_NO_ERROR;
//DEL }


int LDIF_File::Import( void )
{
	assert( m_File != NULL );

	int					Ret = LDIF_NO_ERROR;
	wxString			Dn; 

	fseek( m_File, 0, SEEK_SET );

	Ret = ReadHeader();
	while( Ret == LDIF_NO_ERROR )
	{
		Ret = ReadNextLine();
	}
	return Ret;
}


int LDIF_File::Export( void )
{
	assert( m_File != NULL );

	int				Ret;
	unsigned long	i;
	/*
	*	Header does not depend on what is exported
	*/
	Ret = WriteHeader();

	for( i = 0; i < GetSize(); i++ )
	{
		WriteRecord( i );
	}
	return LDIF_NO_ERROR;
}

int LDIF_File::CreateFile( const wxString & p_FileName )
{

	m_File = fopen( p_FileName.mb_str(), "w" );
	if( m_File == NULL )
	{
		return LDIF_FILE_ERROR;
	}
	return LDIF_NO_ERROR;
}

int LDIF_File::WriteHeader()
{
	assert( m_File != NULL );

	fseek( m_File, 0, SEEK_SET );
	
	fprintf( m_File, "version: 1%c", LDIF_LF );
	
	return LDIF_NO_ERROR ;
}

int LDIF_File::WriteRecord( long p_Index )
{
	assert( m_File != NULL );
	
	wxString	DN;
	berval		**BerValues = NULL;
	DynamicBuffer Value ;
	wxString Name ;

	fprintf( m_File, "%c", LDIF_LF ) ;

	DN = GetDN( p_Index );

	//if( !IsAscii( DN ) )
	if( !IsSafeChar( ( const unsigned char * )DN.c_str(), DN.length(), true ) )
	{
		Base64Encode( Value, ( const unsigned char * )DN.c_str(), DN.length() );
		fprintf( m_File, "%c%cdn:: %s%c%c", LDIF_CR, LDIF_LF , Value.c_str(), 
			LDIF_CR, LDIF_LF );
	}
	else
	{
		fprintf( m_File, "%c%cdn: %s%c%c", LDIF_CR, LDIF_LF , (const char*)DN.mb_str(), LDIF_CR, LDIF_LF );
	}

	//ldap_memfree( DN );
	for( long k = 0; k < GetAttributeCount( p_Index ); k++ )
	{
		BerValues = GetAttributeBinValues( p_Index, k, Name );
		if( BerValues != NULL )
		{
			for( long j = 0 ; BerValues[ j ] != NULL ; j ++ )
			{
				if( IsSafeChar( ( const unsigned char * )BerValues[ j ]->bv_val, 
					BerValues[ j ]->bv_len, true ) )
				{
					fprintf( m_File, "%s: %s%c%c", (const char*)Name.mb_str(), BerValues[ j ]->bv_val, LDIF_CR, LDIF_LF );
				}
				else
				{
					Base64Encode( Value, ( const unsigned char * )BerValues[ j ]->bv_val, 
						BerValues[ j ]->bv_len );
					fprintf( m_File, "%s:: %s%c%c", (const char*)Name.mb_str(), Value.c_str(), LDIF_CR, LDIF_LF );
				}
				
			}
			ldap_value_free_len( BerValues );
		}
	}
	return LDIF_NO_ERROR ;

}

/*bool LDIF_File::IsAscii( const wxString & p_Str )
{
	size_t	Size = p_Str.length();
	bool	Ret = true;

	for( size_t i = 0 ; i < Size ; i ++ )
	{
		if( !isascii( p_Str[ i ].mb_str() ) )
		{
			Ret = false ;
			break;
		}
	}
	return Ret;
}
*/
bool LDIF_File::Must_Encode_Dn( const wxString &p_DN )
{
	/*
	 4)  Any dn or rdn that contains characters other than those
          defined as "SAFE-UTF8-CHAR", or begins with a character other
          than those defined as "SAFE-INIT-UTF8-CHAR", above, MUST be
          base-64 encoded.  Other values MAY be base-64 encoded.  Any
          value that contains characters other than those defined as
          "SAFE-CHAR", or begins with a character other than those
          defined as "SAFE-INIT-CHAR", above, MUST be base-64 encoded.
          Other values MAY be base-64 encoded.
     8)  Values or distinguished names that end with SPACE SHOULD be
          base-64 encoded.

SAFE-STRING              = [SAFE-INIT-CHAR *SAFE-CHAR]
SAFE-INIT-CHAR           = %x01-09 / %x0B-0C / %x0E-1F /
                       %x21-39 / %x3B / %x3D-7F
                       ; any value <= 127 except NUL, LF, CR,
                       ; SPACE, colon (":", ASCII 58 decimal)
                       ; and less-than ("<" , ASCII 60 decimal)

SAFE-CHAR                = %x01-09 / %x0B-0C / %x0E-7F
                           ; any value <= 127 decimal except NUL, LF,
                           ; and CR
SAFE-UTF8-CHAR     = SAFE-CHAR / UTF8-2 / UTF8-3 / UTF8-4 /
                        UTF8-5 / UTF8-6

UTF8-2                   = %xC0-DF UTF8-1

UTF8-3                   = %xE0-EF 2UTF8-1

UTF8-4                   = %xF0-F7 3UTF8-1

UTF8-5                   = %xF8-FB 4UTF8-1

UTF8-6                   = %xFC-FD 5UTF8-1

	*/

	// First test if the string begins with a SAFE-INIT-UTF8-CHAR
	return false;
}

void LDIF_Record::Init()
{
	m_Attributes.clear() ;
	LDIF_RecordAttributes::iterator iter ;
	for( iter = m_Attributes.begin(); iter != m_Attributes.end(); iter++ )
	{
		for( unsigned long i = 0; i < ( *iter ).second.size(); i++ )
		{
			if( ( *iter ).second[ i ] != NULL )
			{
				delete []( *iter ).second[ i ] ;
			}
		}
	}
	m_Attributes.clear();

	LDIF_RecordBervalAttributes::iterator beriter ;
	for( beriter = m_BerAttributes.begin(); beriter != m_BerAttributes.end(); beriter++ )
	{
		for( unsigned long i = 0; i < ( *beriter ).second.size(); i++ )
		{
			if( ( *beriter ).second[ i ] != NULL )
			{
				if( ( *beriter ).second[ i ]->bv_val != NULL )
				{
					delete []( *beriter ).second[ i ]->bv_val ;
				}
				
				delete ( *beriter ).second[ i ] ;
			}
		}
	}
	m_BerAttributes.clear();
	m_Mod_Op = LDAP_MOD_ADD ;
	m_Op = LDAP_OP_ADD ;
	m_Dn = wxT("");
}

int LDIF_File::ReadNextLine()
{
	int				Len ;
	char			FirstChar ;
	//ByteArray		CurrentLine ;
	unsigned int	Pos ;
	wxString		Attribute ;
	wxString		Value ;
	DynamicBuffer	BinValue;
	DynamicBuffer	Decoded;
	bool			IsBinary = false;

	Len = ReadLine();

	if( Len == LDIF_EOF )
	{
		ProcessRecord();
		return Len;
	}

	FirstChar = m_Line[ 0 ] ;
	if( FirstChar == '#' )
	{
		/*
		* This is a comment line
		*/
	}
	else if( strnicmp( m_Line.c_str(), "dn:", 3 ) == 0 )
	{
		/*
		* The previous record shoud be processed and cleaned
		*/

		ProcessRecord();

		Pos = m_Line.find( ':' );
		Value = m_Line.substr( Pos + 1 );

		// 0 is unsigned...
		if( Value.length() > 0 && Value[ 0u ] == ':' )
		{
			
			Value.erase( 0, 1 );
			while( Value[ 0u ] == ' ' && Value.length() > 0 )
			{
				Value.erase( 0, 1 );
			}		
			BinValue = Value ;
			Base64Decode( BinValue, IsBinary, Decoded );
			m_CurrentRecord.m_Dn = wxString( Decoded.c_str(), *wxConvCurrent );
		}
		else if( Value.length() > 0 )
		{
			while( Value[ 0u ] == ' ' && Value.length() > 0 )
			{
				Value.erase( 0, 1 );
			}
			m_CurrentRecord.m_Dn = Value ;
		}
	}
	/*
	else if( strnicmp( m_Line.c_str(), "changetype:", 11 ) == 0 )
	{
		Pos = m_Line.find( ':' );
		if( Pos != m_Line.npos )
		{
			Attribute = m_Line.substr( 0, Pos );
			Value = m_Line.substr( Pos + 1 );

			while( Value[ 0 ] == ' ' && Value.length() > 0 )
			{
				Value.erase( 0, 1 );
			}
			if( strnicmp( Value.c_str(), "add", Value.length() ) == 0 )
			{
				m_CurrentRecord.m_Op = LDAP_OP_ADD ;
			}
			else if( strnicmp( Value.c_str(), "delete", Value.length() ) == 0 )
			{
				m_CurrentRecord.m_Op = LDAP_OP_DEL ;
			}
			else if( strnicmp( Value.c_str(), "modify", Value.length() ) == 0 )
			{
				m_CurrentRecord.m_Op = LDAP_OP_MOD ;
			}
			else if( strnicmp( Value.c_str(), "modrdn", Value.length() ) == 0 )
			{
				m_CurrentRecord.m_Op = LDAP_OP_REN ;
			}
		}
	}
	else if( m_CurrentRecord.m_Op == LDAP_OP_MOD &&
		strnicmp( m_Line.c_str(), "add:", 4 ) == 0 )
	{
		m_CurrentRecord.m_Mod_Op = LDAP_MOD_ADD ;
	}
	else if( m_CurrentRecord.m_Op == LDAP_OP_MOD &&
		strnicmp( m_Line.c_str(), "delete:", 7 ) == 0 )
	{
		m_CurrentRecord.m_Mod_Op = LDAP_MOD_DELETE ;
	}
	else if( m_CurrentRecord.m_Op == LDAP_OP_MOD &&
		strnicmp( m_Line.c_str(), "replace:", 8 ) == 0 )
	{
		m_CurrentRecord.m_Mod_Op = LDAP_MOD_REPLACE ;
	}
	else if( m_CurrentRecord.m_Op == LDAP_OP_MOD && FirstChar == '-' )
	{
		ProcessRecord();
	}
	*/
	else
	{
		Pos = m_Line.find( ':' );
		if( Pos != m_Line.npos() )
		{
			Attribute = m_Line.substr( 0, Pos );
			Value = m_Line.substr( Pos + 1 );
			if( Value[ 0u ] == ':' )
			{
				Value.erase( 0, 1 );
				while( Value[ 0u ] == ' ' && Value.length() > 0 )
				{
					Value.erase( 0, 1 );
				}
				BinValue = Value;
				Base64Decode( BinValue, IsBinary, Decoded ) ;
				if( IsBinary )
				{
					m_CurrentRecord.AddBinValue( Attribute, Decoded );
				}
				else
				{
					m_CurrentRecord.AddValue( Attribute, wxString( Decoded.c_str(), *wxConvCurrent ) );
				}
			}
			else
			{
				while( Value[ 0u ] == ' ' && Value.length() > 0 )
				{
					Value.erase( 0, 1 );
				}
				m_CurrentRecord.AddValue( Attribute, Value );
			}
		}
	}
	return LDIF_NO_ERROR ;
}

int LDIF_File::ProcessRecord()
{
	int Ret ;
	LDAPMod ** Mods = NULL;

	if( m_CurrentRecord.m_Dn.length() == 0 )
	{
		m_CurrentRecord.Init();
		return LDIF_BAD_RECORD ;
	}
	switch( m_CurrentRecord.m_Op )
	{
	case LDAP_OP_MOD :
		Mods = m_CurrentRecord.Get_LDAPMod_Array();
		if( Mods != NULL )
		{
			Ret = ldap_modify_s( m_LDAP, m_CurrentRecord.m_Dn.mb_str(), Mods );
			m_CurrentRecord.Free_LDAPMod_Array( Mods );
		}
		m_CurrentRecord.Init();
		m_CurrentRecord.m_Op = LDAP_OP_MOD ;
		break;
	case LDAP_OP_DEL :
		Ret = ldap_delete_s( m_LDAP, m_CurrentRecord.m_Dn.mb_str() );
		m_CurrentRecord.Init();
		break;
	case LDAP_OP_ADD :
		Mods = m_CurrentRecord.Get_LDAPMod_Array();
		if( Mods != NULL )
		{
			Ret = ldap_add_s( m_LDAP, m_CurrentRecord.m_Dn.mb_str(), Mods );
			if( Ret != LDAP_SUCCESS )
			{
				wxString ErrMsg = wxString( ldap_err2string( Ret ), *wxConvCurrent );
				::wxMessageBox( ErrMsg, _( "LDAP Error" ), 
					wxOK | wxCENTER | wxICON_EXCLAMATION );
			}
			m_CurrentRecord.Free_LDAPMod_Array( Mods );
		}
		m_CurrentRecord.Init();
		m_CurrentRecord.m_Op = LDAP_OP_ADD ;
		break;
	case LDAP_OP_REN :
		break;
	}
	return 0 ;
}

LDAPMod ** LDIF_Record::Get_LDAPMod_Array()
{
	LDAPMod									**RetMods = NULL ;
	LDIF_RecordAttributes::iterator			iter ;
	LDIF_RecordBervalAttributes::iterator	beriter ;
	long									i = 0 ;
	long									NAtts;

	NAtts = m_Attributes.size() + m_BerAttributes.size() + 1 ;
	RetMods = new LDAPMod *[ NAtts ] ;
	if( RetMods == NULL )
	{
		return NULL ;
	}
	for( i = 0; i < NAtts; i ++ )
	{
		RetMods[ i ] = NULL ;
	}
	i = 0;
	for( iter = m_Attributes.begin(); iter != m_Attributes.end(); iter ++ )
	{
		/*
		*  Parsing the text attributes
		*/
		RetMods[ i ] = NULL ;
		RetMods[ i ] = new LDAPMod ;
		if( RetMods[ i ] == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}
		
		//RetMods[ i ]->mod_op = m_Mod_Op ;
		RetMods[ i ]->mod_op = m_Mod_Op ;

		/*
		*	(*iter) is a pair: first = name of the attribute,
		*	second = vector of attribute value as string
		*/
		long	mod_type_size = ( *iter ).first.length() + 1 ;
		RetMods[ i ]->mod_type = new char[ mod_type_size ] ;
		if( RetMods[ i ]->mod_type == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}
		memset( RetMods[ i ]->mod_type, '\0', mod_type_size );
		strcpy( RetMods[ i ]->mod_type, ( *iter ).first.mb_str() );

		long NValues ;

		NValues = ( *iter ).second.size() + 1 ;
		
		RetMods[ i ]->mod_values = new char *[ NValues ];
		if( RetMods[ i ]->mod_values == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}
		long j;
		for( j = 0 ; j < NValues; j++ )
		{
			RetMods[ i ]->mod_values[ j ] = NULL ;
		}
		for( j = 0 ; j < ( NValues - 1 ); j ++ )
		{
			RetMods[ i ]->mod_values[ j ] = ( *iter ).second[ j ] ;
			/*
			long mod_values_size = ( *iter ).second[ j ].length() + 1 ;

			RetMods[ i ]->mod_values[ j ] = new char[ mod_values_size ] ;
			if( RetMods[ i ]->mod_values[ j ] == NULL )
			{
				Free_LDAPMod_Array( RetMods );
				return NULL ;
			}
			memset( RetMods[ i ]->mod_values[ j ], '\0', mod_values_size );
			strcpy( RetMods[ i ]->mod_values[ j ], ( *iter ).second[ j ].c_str() ) ;
			*/
		}
		i++ ;
	}
	for( beriter = m_BerAttributes.begin(); beriter != m_BerAttributes.end(); beriter ++ )
	{
		RetMods[ i ] = NULL ;
		RetMods[ i ] = new LDAPMod ;
		if( RetMods[ i ] == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}
		
		RetMods[ i ]->mod_op = m_Mod_Op | LDAP_MOD_BVALUES ;

		long	mod_type_size = ( *beriter ).first.length() + 1 ;
		RetMods[ i ]->mod_type = new char[ mod_type_size ] ;
		if( RetMods[ i ]->mod_type == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}
		memset( RetMods[ i ]->mod_type, '\0', mod_type_size );
		strcpy( RetMods[ i ]->mod_type, ( *beriter ).first.mb_str() );

		long NValues ;

		NValues = ( *beriter ).second.size() + 1 ;
		
		RetMods[ i ]->mod_bvalues = new berval *[ NValues ];
		if( RetMods[ i ] == NULL )
		{
			Free_LDAPMod_Array( RetMods );
			return NULL ;
		}

		long j;
		
		for( j = 0 ; j < NValues; j++ )
		{
			RetMods[ i ]->mod_bvalues[ j ] = NULL ;
		}
		for( j = 0 ; j < ( NValues - 1 ); j ++ )
		{
			RetMods[ i ]->mod_bvalues[ j ] = ( *beriter ).second[ j ] ;
			/*
			RetMods[ i ]->mod_bvalues[ j ] = new berval ;
			if( RetMods[ i ]->mod_bvalues[ j ] == NULL )
			{
				Free_LDAPMod_Array( RetMods );
				return NULL ;
			}
			RetMods[ i ]->mod_bvalues[ j ]->bv_val = NULL ;
			RetMods[ i ]->mod_bvalues[ j ]->bv_val = new char[ ( *beriter ).second[ j ]->bv_len ] ;
			if( RetMods[ i ]->mod_bvalues[ j ]->bv_val != NULL )
			{
				memcpy( RetMods[ i ]->mod_bvalues[ j ]->bv_val, ( *beriter ).second[ j ]->bv_val,
					( *beriter ).second[ j ]->bv_len ) ;
				RetMods[ i ]->mod_bvalues[ j ]->bv_len = ( *beriter ).second[ j ]->bv_len ;
			}
			*/
		}
		i ++ ;
	}
	return RetMods;
}

void LDIF_Record::Free_LDAPMod_Array(LDAPMod **p_Mods)
{
	if( p_Mods == NULL )
		return;

	for( long i = 0; p_Mods[ i ] != NULL; i++ )
	{
		if( p_Mods[ i ]->mod_type != NULL )
		{
			delete []p_Mods[ i ]->mod_type ;
		}
		if( p_Mods[ i ]->mod_values != NULL )
		{
			/*
			for( long j = 0; p_Mods[ i ]->mod_values[ j ] != NULL ; j ++ )
			{
				delete []p_Mods[ i ]->mod_values[ j ] ;
			}
			*/
			delete []p_Mods[ i ]->mod_values ;
		}
		delete p_Mods[ i ];
	}

	delete []p_Mods ;
	p_Mods = NULL ;
}

void LDIF_Record::AddValue( const wxString &p_Name, const wxString &p_Value )
{
	char	*NewValue = NULL ;
	NewValue = new char[ p_Value.length() + 1 ];
	if( NewValue == NULL )
	{
		return ;
	}
	strcpy( NewValue, p_Value.mb_str() );
	m_Attributes[ p_Name ].push_back( NewValue );	
}

void LDIF_Record::AddBinValue(const wxString &p_Name, const DynamicBuffer &p_Value)
{
	berval	*Ber = NULL ;

	Ber = new berval ;
	if( Ber == NULL )
	{
		return;
	}

	Ber->bv_len = p_Value.length();
	Ber->bv_val = new char[ Ber->bv_len ] ;
	if( Ber->bv_val == NULL )
	{
		return ;
	}
	
	memset( Ber->bv_val, '\0', Ber->bv_len );
	memcpy( Ber->bv_val, p_Value.c_str(), p_Value.length() );
	m_BerAttributes[ p_Name ].push_back( Ber );

}
